#!/usr/local/bin/ruby -w
#
# vim:sw=2 ts=8:et sta:fdm=marker
#
#
# Copyright (c) 2005 Ariff Abdullah
#        (skywizard@MyBSD.org.my) All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#        $MyBSD$
#
# Date: Wed Nov 23 21:47:41 MYT 2005
#   OS: FreeBSD osaka.MyBSD.org.my 6.0-RELEASE i386
#

# {{{ ACPIIO
# {{{ module IOCCOM
module IOCCOM
  PARM_MASK = 0x1fff
  VOID      = 0x20000000
  OUT       = 0x40000000
  IN        = 0x80000000
  INOUT     = IN|OUT
  DIRMASK   = 0xe0000000
  def PARM_LEN(x)
    (x >> 16) & PARM_MASK
  end
  def BASECMD(x)
    x & ~(PARM_MASK << 16)
  end
  def GROUP(x)
    (x >> 8) & 0xff
  end
  def _IOC(inout, group, num, len)
    inout | ((len & PARM_MASK) << 16) | ((group.is_a?(String) ? group[0] : group) << 8) | num
  end
  def _IO(g, n)
    _IOC(VOID, g, n, 0)
  end
  def _IOR(g, n, t)
    _IOC(OUT, g, n, t)
  end
  def _IOW(g, n, t)
    _IOC(IN, g, n, t)
  end
  def _IOWR(g, n, t)
    _IOC(INOUT, g, n, t)
  end
  module_function :PARM_LEN, :BASECMD, :GROUP
  module_function :_IOC, :_IO, :_IOR, :_IOW, :_IOWR
end
# }}}

# {{{ module ACPIIO_CONSTANTS
module ACPIIO_CONSTANTS
  CMBAT_MAXSTRLEN = 32
  BATT_STAT_DISCHARG    = 0x0001
  BATT_STAT_CHARGING    = 0x0002
  BATT_STAT_CRITICAL    = 0x0004
  BATT_STAT_NOT_PRESENT = 0x0007
  BATT_STAT_MAX         = 0x0007
  BATTERY_ALL_UNITS     = -1
  BATT_UNKNOWN          = 0xffffffff
  BIF_UNITS_MW = 0
  BIF_UNITS_MA = 1
  # XXX too lazy to calculate
  SIZEOF_UNION_ACPI_BATTERY_IOCTL_ARG = 164
  SIZEOF_STRUCT_ACPI_BATTINFO         = 16
  SIZEOF_STRUCT_ACPI_BIF              = 164
  SIZEOF_STRUCT_ACPI_BST              = 16
  SIZEOF_INT                          = 4
  ACPIDEV = '/dev/acpi'
end
# }}}

# {{{ module ACPIIO
module ACPIIO
  class << self
    include IOCCOM
  end
  include ACPIIO_CONSTANTS
  SETSLPSTATE       = _IOW('P', 3, SIZEOF_INT)
  BATT_GET_UNITS    = _IOR('B', 0x01, SIZEOF_INT)
  BATT_GET_BATTINFO = _IOWR('B', 0x03, SIZEOF_UNION_ACPI_BATTERY_IOCTL_ARG)
  BATT_GET_BIF      = _IOWR('B', 0x10, SIZEOF_UNION_ACPI_BATTERY_IOCTL_ARG)
  BATT_GET_BST      = _IOWR('B', 0x11, SIZEOF_UNION_ACPI_BATTERY_IOCTL_ARG)
  ACAD_GET_STATUS   = _IOR('A', 1, SIZEOF_INT)
end
# }}}

# {{{ class Acpiio
class Acpiio
  include ACPIIO
  def initialize(dev = ACPIDEV)
    @fd = File.open(dev, 'r')
  end
  def close
    @fd.close()
  end
  def closed?
    @fd.closed?
  end
  def get_bif(num)
    buf = _construct_arg(num)
    @fd.ioctl(BATT_GET_BIF, buf)
    ret = buf.unpack('IIIIIIIIIa32a32a32a32')
    ret.collect! do |val|
      val.rstrip! if val.is_a?(String)
      val
    end
    ret
  end
  def get_battinfo(num)
    buf = _construct_arg(num)
    @fd.ioctl(BATT_GET_BATTINFO, buf)
    ret = buf.unpack('iiii')
    ret
  end
  def get_bst(num)
    buf = _construct_arg(num)
    @fd.ioctl(BATT_GET_BST, buf)
    ret = buf.unpack('IIII')
    ret
  end
  def acline
    buf = [0].pack('i')
    @fd.ioctl(ACAD_GET_STATUS, buf)
    ret = buf.unpack('i').first()
    ret
  end
  private
  def _construct_arg(num)
    buf = [num].pack('i')
    buf << "\000"*(SIZEOF_UNION_ACPI_BATTERY_IOCTL_ARG - buf.size())
  end
end
# }}}

# {{{ class AcpiioBIF
class AcpiioBIF
  include ACPIIO_CONSTANTS
  FUNCS = %w[units dcap lfcap btech dvol wcap lcap gra1 gra2
                                      model serial type oeminfo]
  class << self
      FUNCS.each_with_index do |func, idx|
      module_eval %{
        def #{func}(data)
          data[#{idx}]
        end
      }
    end
    def pwr_units(data)
      if data[0] == self::BIF_UNITS_MW
        'mW'
      else
        'mA'
      end
    end
  end
  def initialize(acpi, num = 0)
    @acpi = acpi
    @battery_num = num
    update()
  end
  def update
    @data = @acpi.get_bif(@battery_num)
  end
  FUNCS.each do |func|
    class_eval %{
      def #{func}
        self.class.#{func}(@data)
      end
    }
  end
  def pwr_units
    self.class.pwr_units(@data)
  end
end
# }}}

# {{{ class AcpiioBATTINFO
class AcpiioBATTINFO
  include ACPIIO_CONSTANTS
  FUNCS = %w[cap min state rate]
  class << self
      FUNCS.each_with_index do |func, idx|
      module_eval %{
        def #{func}(data)
          data[#{idx}]
        end
      }
    end
  end
  def initialize(acpi, num = 0)
    @acpi = acpi
    @battery_num = num
    update()
  end
  def update
    @data = @acpi.get_battinfo(@battery_num)
  end
  FUNCS.each do |func|
    class_eval %{
      def #{func}
        self.class.#{func}(@data)
      end
    }
  end
end
# }}}

# {{{ class AcpiioBST
class AcpiioBST
  include ACPIIO_CONSTANTS
  FUNCS = %w[state rate cap volt]
  class << self
      FUNCS.each_with_index do |func, idx|
      module_eval %{
        def #{func}(data)
          data[#{idx}]
        end
      }
    end
  end
  def initialize(acpi, num = 0)
    @acpi = acpi
    @battery_num = num
    update()
  end
  def update
    @data = @acpi.get_bst(@battery_num)
  end
  FUNCS.each do |func|
    class_eval %{
      def #{func}
        self.class.#{func}(@data)
      end
    }
  end
end
# }}}

# {{{ class Acpiconf < Acpiio
class Acpiconf < Acpiio
  def initialize(dev = ACPIDEV, num = 0)
    super(dev)
    self.battery_unit = num
  end
  def battery_unit
    @battery_unit ||= nil
    @battery_unit
  end
  def battery_unit=(num)
    @battery_unit ||= nil
    @bif ||= nil
    if @battery_unit != num || !@bif.is_a?(AcpiioBIF)
      @battery_unit = num
      @bif = AcpiioBIF.new(self, @battery_unit)
    else
      @bif.update()
    end
    @battery_unit
  end
  def bif
    @bif ||= nil
    @bif
  end
  def battinfo
    AcpiioBATTINFO.new(self, @battery_unit)
  end
  def bst
    AcpiioBST.new(self, @battery_unit)
  end
  def acline?
    ac = acline()
    if ac == 0
      false
    elsif ac == 1
      true
    else
      raise Errno::EINVAL, "Invalid acline value - #{ac}"
    end
  end
  # {{{ def info
  def info
    ret = ''
    bif = bif()
    pwr_units = bif.pwr_units()
    dcap = bif.dcap()
    if dcap == BATT_UNKNOWN
      ret << sprintf("Design capacity:\tunknown\n")
    else
      ret << sprintf("Design capacity:\t%d %sh\n", dcap, pwr_units)
    end
    lfcap = bif.lfcap()
    if lfcap == BATT_UNKNOWN
      ret << sprintf("Last full capacity:\tunknown\n");
    else
      ret << sprintf("Last full capacity:\t%d %sh\n", lfcap, pwr_units)
    end
    if bif.btech() == 0
      ret << sprintf("Technology:\t\tprimary (non-rechargeable)\n")
    else
      ret << sprintf("Technology:\t\tsecondary (rechargeable)\n")
    end
    dvol = bif.dvol()
    if dvol == BATT_UNKNOWN
      ret << sprintf("Design voltage:\t\tunknown\n")
    else
      ret << sprintf("Design voltage:\t\t%d mV\n", dvol)
    end
    ret << sprintf("Capacity (warn):\t%d %sh\n", bif.wcap(), pwr_units);
    ret << sprintf("Capacity (low):\t\t%d %sh\n", bif.lcap(), pwr_units);
    ret << sprintf("Low/warn granularity:\t%d %sh\n", bif.gra1(), pwr_units);
    ret << sprintf("Warn/full granularity:\t%d %sh\n", bif.gra2(), pwr_units);
    ret << sprintf("Model number:\t\t%s\n", bif.model());
    ret << sprintf("Serial number:\t\t%s\n", bif.serial());
    ret << sprintf("Type:\t\t\t%s\n", bif.type());
    ret << sprintf("OEM info:\t\t%s\n", bif.oeminfo());
    # XXX BATTINFO
    binfo = battinfo()
    state = binfo.state()
    if state != BATT_STAT_NOT_PRESENT
      ret << sprintf("State:\t\t\t")
      ret << sprintf('high ') if state == 0
      ret << sprintf('critical ') if (state & BATT_STAT_CRITICAL) != 0
      ret << sprintf('discharging ') if (state & BATT_STAT_DISCHARG) != 0
      ret << sprintf('charging ') if (state & BATT_STAT_CHARGING) != 0
      ret << sprintf("\n")
      cap = binfo.cap()
      if cap == -1
        ret << sprintf("Remaining capacity:\tunknown\n")
      else
        ret << sprintf("Remaining capacity:\t%d%%\n", cap)
      end
      min = binfo.min()
      if min == -1
        ret << sprintf("Remaining time:\t\tunknown\n")
      else
        ret << sprintf("Remaining time:\t\t%d:%02d\n", min / 60, min % 60)
      end
      rate = binfo.rate()
      if rate == -1
        ret << sprintf("Present rate:\t\tunknown\n")
      else
        ret << sprintf("Present rate:\t\t%d %s\n", rate, pwr_units)
      end
    #else
    #  ret << sprintf("BATTINFO state:\t\tnot present\n")
    end
    # BST
    bst = bst()
    state = bst.state()
    if state != BATT_STAT_NOT_PRESENT
      if binfo.state() == BATT_STAT_NOT_PRESENT
        ret << sprintf("State:\t\t\t")
        ret << sprintf('high ') if state == 0
        ret << sprintf('critical ') if (state & BATT_STAT_CRITICAL) != 0
        ret << sprintf('discharging ') if (state & BATT_STAT_DISCHARG) != 0
        ret << sprintf('charging ') if (state & BATT_STAT_CHARGING) != 0
        ret << sprintf("\n")
        rate = bst.rate()
        if rate == BATT_UNKNOWN
          ret << sprintf("Present rate:\t\tunknown\n")
        else
          ret << sprintf("Present rate:\t\t%d %s\n", rate, pwr_units)
        end
        cap = bst.cap()
        if cap == BATT_UNKNOWN
          ret << sprintf("Remaining capacity:\tunknown\n")
        else
          ret << sprintf("Remaining capacity:\t%d%%\n", (100 * cap) / bif.lfcap())
        end
      end
      volt = bst.volt()
      if volt == BATT_UNKNOWN
        ret << sprintf("Voltage:\t\tunknown\n")
      else
        ret << sprintf("Voltage:\t\t%d mV\n", volt)
      end
    #else
    #  ret << sprintf("BST state:\t\tnot present\n")
    end
    if acline?
      ret << sprintf("AC line:\t\ton-line\n")
    else
      ret << sprintf("AC line:\t\toff-line\n")
    end
    if binfo.state() == BATT_STAT_NOT_PRESENT &&
          bst.state() == BATT_STAT_NOT_PRESENT
        ret << sprintf("State:\t\t\tnot present\n")
    end
    ret
  end
  # }}}
  def info_a
    info.split(/\n/).collect do |v|
      v.split(/:\t+/, 2)
    end
  end
end
# }}}
# }}}

if __FILE__ == $0
  begin
    acpi = Acpiconf.new()
    acpi.battery_unit = 0
    print acpi.info()
  ensure
    acpi.close() if acpi && !acpi.closed?
  end
end
